"""Test the handlers."""

import json
import shutil
from pathlib import Path
from typing import Any, Callable

import pytest
import tornado


async def test_get_file_no_path(jp_fetch: Callable[..., Any]) -> None:
    """Test the GET file handler without a path.
    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    """
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch("waldiez", "files")
    assert exc_info.value.code == 400


async def test_get_file_not_fount(jp_fetch: Callable[..., Any]) -> None:
    """Test the GET file handler with a non-existing path.
    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    """
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch("waldiez", "files", params={"path": "not_found.txt"})
    assert exc_info.value.code == 404


async def test_get_file(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
) -> None:
    """Test the GET file handler with a valid path.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    """
    # first create a file in the root directory
    file_path = jp_root_dir / "example.waldiez"
    file_path.write_text("test")
    response = await jp_fetch(
        "waldiez",
        "files",
        params={"path": "example.waldiez"},
    )
    assert response.code == 200
    expected = {"path": str(file_path)}
    assert response.body == json.dumps(expected).encode("utf-8")
    file_path.unlink()


async def test_export_to_py(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
    data_dir: Path,
) -> None:
    """Test exporting a .waldiez file to a .py file.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    data_dir : Path
        The data directory.
    """
    waldiez_path = data_dir / "flow.waldiez"
    shutil.copy(data_dir / "flow.waldiez", jp_root_dir / "flow.waldiez")
    response = await jp_fetch(
        "waldiez",
        "files",
        method="POST",
        body=json.dumps(
            {
                "files": [str(waldiez_path)],
                "extension": "py",
            }
        ),
    )
    assert response.code == 200
    relative_to_cwd = waldiez_path.relative_to(Path.cwd())
    expected = {"files": [str(relative_to_cwd.with_suffix(".py"))]}
    # expected = {"files": [str(waldiez_path)]}
    assert response.body == json.dumps(expected).encode("utf-8")
    assert (data_dir / "flow.py").exists()
    (data_dir / "flow.py").unlink()
    (jp_root_dir / "flow.waldiez").unlink()
    if (data_dir / "waldiez_api_keys.py").exists():
        (data_dir / "waldiez_api_keys.py").unlink()


async def test_export_to_ipynb(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
    data_dir: Path,
) -> None:
    """Test exporting a .waldiez file to a .ipynb file.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    data_dir : Path
        The data directory.
    """
    waldiez_path = data_dir / "flow.waldiez"
    shutil.copy(data_dir / "flow.waldiez", jp_root_dir / "flow.waldiez")
    response = await jp_fetch(
        "waldiez",
        "files",
        method="POST",
        body=json.dumps(
            {
                "files": [str(waldiez_path)],
                "extension": "ipynb",
            }
        ),
    )
    assert response.code == 200
    relative_to_cwd = waldiez_path.relative_to(Path.cwd())
    expected = {"files": [str(relative_to_cwd.with_suffix(".ipynb"))]}
    # expected = {"files": [str(waldiez_path)]}
    assert response.body == json.dumps(expected).encode("utf-8")
    assert (data_dir / "flow.ipynb").exists()
    (data_dir / "flow.ipynb").unlink()
    (jp_root_dir / "flow.waldiez").unlink()
    if (data_dir / "waldiez_api_keys.py").exists():
        (data_dir / "waldiez_api_keys.py").unlink()


async def test_export_to_invalid_extension(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
    data_dir: Path,
) -> None:
    """Test exporting a .waldiez file with an invalid extension.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    data_dir : Path
        The data directory.
    """
    waldiez_path = data_dir / "flow.waldiez"
    shutil.copy(data_dir / "flow.waldiez", jp_root_dir / "flow.waldiez")
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch(
            "waldiez",
            "files",
            method="POST",
            body=json.dumps(
                {
                    "files": [str(waldiez_path)],
                    "extension": "invalid",
                }
            ),
        )
    assert exc_info.value.code == 400
    assert exc_info.value.response
    (jp_root_dir / "flow.waldiez").unlink()
    if (data_dir / "waldiez_api_keys.py").exists():
        (data_dir / "waldiez_api_keys.py").unlink()


async def test_export_from_invalid_extension(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
    data_dir: Path,
) -> None:
    """Test exporting a file with an invalid extension.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    data_dir : Path
        The data directory.
    """
    waldiez_path = data_dir / "flow.invalid"
    shutil.copy(data_dir / "flow.waldiez", jp_root_dir / "flow.invalid")
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch(
            "waldiez",
            "files",
            method="POST",
            body=json.dumps(
                {
                    "files": [str(waldiez_path)],
                    "extension": "py",
                }
            ),
        )
    assert exc_info.value.code == 400



# type: ignore
"""Jupyter server extension for Waldiez."""

from typing import Dict, List

from jupyter_server.serverapp import ServerApp

# pylint: disable=import-error
# pyright: reportMissingImports=false
try:
    from ._version import __version__  # noqa
except ImportError:
    # Fallback when using the package in dev mode without installing
    # in editable mode with pip. It is highly recommended to install
    # the package from a stable release or in editable mode:
    # https://pip.pypa.io/en/stable/topics/local-project-installs/#editable-installs
    import warnings

    warnings.warn("Importing 'waldiez_jupyter' outside a proper installation.")
    __version__ = "dev"


from .handlers import setup_handlers


def _jupyter_labextension_paths() -> List[Dict[str, str]]:
    return [{"src": "labextension", "dest": "@waldiez/jupyter"}]


def _jupyter_server_extension_points() -> List[Dict[str, str]]:
    return [{"module": "waldiez_jupyter"}]


def _load_jupyter_server_extension(server_app: ServerApp) -> None:
    """Register the API handler to receive HTTP requests from the frontend.

    Parameters
    ----------
    server_app: ServerApp
        JupyterLab application instance
    """
    setup_handlers(server_app.web_app)
    name = "@waldiez/jupyter"
    server_app.log.info(f"Registered {name} server extension")



"""Handle requests to handle .waldiez files.

GET:
    Get the actual path of a .waldiez file:
    On the frontend, the path is relative to server's setting notebook_dir.
    We need to get the actual path of the file on the server.

POST:
    Export the .waldiez files to the specified extension.
    The request should contain the list of files to export and the
        target extension to export to (either "py" or "ipynb").
    The response will contain the list of files that were exported.
"""

import json
import logging
import os
from pathlib import Path
from typing import Awaitable, List, Tuple

from jupyter_server.base.handlers import APIHandler
from tornado.web import HTTPError, authenticated
from waldiez import WaldiezExporter

# pylint: disable=broad-except


class FilesHandler(APIHandler):
    """Files handler to get or export .waldiez files."""

    @authenticated
    def data_received(self, chunk: bytes) -> Awaitable[None] | None:
        """Just to make the linter happy.

        avoid: Pylint W0223:abstract-method

        Parameters
        ----------
        chunk : bytes
            The chunk of data received.

        Returns
        -------
        Awaitable[None] | None
            The data handling coroutine.
        """
        return super().data_received(chunk)  # pragma: no cover

    @authenticated
    async def get(self) -> None:
        """Handle a GET request.

        Raises
        ------
        HTTPError
            If the request data is invalid.

        Get the actual path of a .waldiez file.
        The request should contain the path of a file.

        Example URL:
        /waldiez/files?path=relative/to/example.waldiez
        """
        file_path_art = self.get_query_argument("path", None)
        if not file_path_art:
            raise HTTPError(400, reason="No path in request")
        try:
            file_path = self._get_file_path(file_path_art)
        except FileNotFoundError as error:
            raise HTTPError(404, reason=str(error)) from error
        self.finish(json.dumps({"path": str(file_path)}))

    @authenticated
    async def post(self) -> None:
        """Handle a POST request.

        Raises
        ------
        HTTPError
            If the request data is invalid.

        The request should contain the list of files to export and the
        target extension to export to (either "py" or "ipynb").

        Example JSON body:
        {
            "files": ["file1.waldiez", "file2.waldiez"],
            "extension": "py"
        }
        """
        try:
            files, target_extension = self._gather_post_data()
        except HTTPError as error:
            raise error
        if not files:
            raise HTTPError(400, reason="No valid files in the request")
        results = _handle_export(files, target_extension)
        self.log.info("Exported: %s", results)
        self.finish(json.dumps({"files": results}))

    def _gather_post_data(self) -> Tuple[List[str], str]:
        """Gather the data from the POST request.

        Returns
        -------
        Tuple[List[str], str]
            The list of files and the target extension.

        Raises
        ------
        ValueError
            If the request data is invalid
        """
        input_data = self.get_json_body()
        if not input_data:
            raise HTTPError(400, reason="No data in request")
        files = input_data.get("files", [])
        target_extension = input_data.get("extension", "")
        if target_extension not in ("py", "ipynb"):
            raise HTTPError(400, reason="Invalid extension")
        if not isinstance(files, list) or not files:
            raise HTTPError(400, reason="No files in request")
        try:
            return self._get_file_paths(files), target_extension
        except BaseException as error:
            raise HTTPError(400, reason="Error getting file paths") from error

    def _get_file_path(self, file: str) -> Path:
        """Get the actual path of the file.

        Parameters
        ----------
        file : str
            The file path.

        Returns
        -------
        str
            The actual path of the file.

        Raises
        ------
        FileNotFoundError
            If the file is not found.
        """
        if os.path.exists(file) and os.path.isfile(file):
            return Path(os.path.abspath(file))
        joined = os.path.join(self.contents_manager.root_dir, file)
        self.log.error("Joined: %s", joined)
        if os.path.exists(joined) and os.path.isfile(joined):
            return Path(os.path.abspath(joined))
        raise FileNotFoundError(f"File not found: {file}")

    def _get_file_paths(self, files: List[str]) -> List[str]:
        """Get the actual paths of the files.

        Parameters
        ----------
        files : List[str]
            The list of files.
        Returns
        -------
        List[str]
            The list of actual paths of the files.
        """
        file_paths = []
        for file in files:
            if not isinstance(file, str) or not file.endswith(".waldiez"):
                continue
            try:
                actual_file_path = self._get_file_path(file)
            except BaseException as error:
                self.log.error("Error getting file path: %s", error)
                continue
            file_paths.append(str(actual_file_path))
        return file_paths


def _relative_to_cwd(file_path: Path) -> str:
    """Get the relative path to the current working directory.

    Parameters
    ----------
    file_path : Path
        The path to the file.

    Returns
    -------
    str
        The relative path to the current working directory.
    """
    file_path_str = str(file_path).replace(str(Path.cwd().resolve()), "")
    if file_path_str.startswith(os.path.sep):
        file_path_str = file_path_str[len(os.path.sep) :]
    return file_path_str


def _handle_export(files: list[str], target_extension: str) -> List[str]:
    """Handle the export.

    Parameters
    ----------
    files : list[str]
        The list of files to export.
    target_extension : str
        The target extension to export to.

    Returns
    -------
    List[str]
        The list of files that were exported.
    """
    file_paths = []
    for file in files:
        file_path = Path(file).resolve()
        try:
            exporter = WaldiezExporter.load(file_path)
        except BaseException as error:  # pragma: no cover
            logging.debug("Error loading file: %s", error)
            continue
        if target_extension == "py":
            file_path = file_path.with_suffix(".py")
            try:
                exporter.export(file_path, force=True)
            except BaseException as error:  # pragma: no cover
                logging.debug("Error exporting to .py: %s", error)
                continue
            to_cwd = _relative_to_cwd(file_path)
            file_paths.append(to_cwd)
        elif target_extension == "ipynb":
            file_path = file_path.with_suffix(".ipynb")
            try:
                exporter.export(file_path, force=True)
            except BaseException as error:  # pragma: no cover
                logging.debug("Error exporting to .ipynb: %s", error)
                continue
            exporter.export(file_path, force=True)
            to_cwd = _relative_to_cwd(file_path)
            file_paths.append(to_cwd)
    return file_paths



"""Generate requirements txt files from pyproject.toml."""

# pylint: disable=line-too-long

import os
import shutil
import subprocess  # nosemgrep # nosec
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent.parent
EXTRAS = ["dev", "test"]


def _write_all_dot_txt() -> None:
    """Generate requirements/all.txt with references to all requirements."""
    items = EXTRAS + ["main"]
    items.sort()
    with open("requirements/all.txt", "w", encoding="utf-8") as file:
        for item in items:
            file.write(f"-r {item}.txt\n")


def _ensure_uv() -> None:
    """Ensure that the `uv` tool is installed."""
    if not shutil.which("uv"):
        subprocess.run(["python", "-m", "pip", "install", "uv"], check=True)  # nosemgrep # nosec
    try:
        subprocess.run(["uv", "--version"], check=True, stdout=subprocess.DEVNULL)  # nosemgrep # nosec
    except subprocess.CalledProcessError:
        print("Failed to run `uv`.")
        print("Please make sure that the `uv` tool is installed.")
        print("You can install it using `python -m pip install uv`.")
        sys.exit(1)


def main() -> None:
    """Generate requirements/*txt files from pyproject.toml."""
    _ensure_uv()
    subprocess.run(  # nosemgrep # nosec
        [
            "uv",
            "pip",
            "compile",
            "pyproject.toml",
            "--output-file",
            "requirements/main.txt",
        ],
        check=True,
        cwd=str(ROOT_DIR),
    )
    for extra in EXTRAS:
        subprocess.run(  # nosemgrep # nosec
            [
                "uv",
                "pip",
                "compile",
                "pyproject.toml",
                "--output-file",
                f"requirements/{extra}.txt",
                f"--extra={extra}",
                "--no-deps",
            ],
            check=True,
            cwd=str(ROOT_DIR),
        )
    _write_all_dot_txt()
    print("Done. Generated:")
    for file in os.listdir("requirements"):
        print(f"  - {file}")


if __name__ == "__main__":
    main()



"""Initialize the development environment."""

import os
import shutil
import subprocess  # nosemgrep # nosec
import sys
from pathlib import Path
from typing import List, Optional, Union

ROOT_DIR = Path(__file__).resolve().parents[2]
VENV_DIR = ROOT_DIR / ".venv"
REQUIREMENTS_FILE = ROOT_DIR / "requirements" / "all.txt"

BIN_DIR = VENV_DIR / "Scripts" if sys.platform == "win32" else VENV_DIR / "bin"
DOR_EXE = ".exe" if sys.platform == "win32" else ""
PYTHON_COMMAND = str(BIN_DIR / f"python{DOR_EXE}")
JUPYTER_COMMAND = str(BIN_DIR / f"jupyter{DOR_EXE}")
PIP_INSTALL = [PYTHON_COMMAND, "-m", "pip", "install"]


def run_command(cmd: Union[List[str], str], cwd: Optional[str] = None) -> None:
    """Run a command.

    Parameters
    ----------
    cmd : Union[List[str], str]
        The command to run.

    cwd : Optional[str], optional
        The current working directory, by default None.
    """
    if isinstance(cmd, str):
        cmd = cmd.split()
    print(f"Running command: {cmd}")
    if cwd is None:
        cwd = str(ROOT_DIR)
    env = os.environ
    subprocess.run(cmd, env=env, cwd=cwd, check=True)  # nosemgrep # nosec


def create_venv() -> None:
    """Create a virtual environment."""
    if BIN_DIR.is_dir() and os.access(PYTHON_COMMAND, os.X_OK):
        # pylint: disable=inconsistent-quotes
        os.environ["PATH"] = f"{BIN_DIR}{os.pathsep}{os.environ['PATH']}"
        return
    if BIN_DIR.is_dir():
        shutil.rmtree(BIN_DIR)
    print("Creating virtual environment...")
    run_command([sys.executable, "-m", "venv", str(VENV_DIR)])
    # pylint: disable=inconsistent-quotes
    os.environ["PATH"] = f"{BIN_DIR}{os.pathsep}{os.environ['PATH']}"


def pip_upgrade_base() -> None:
    """Upgrade pip setuptools and wheel."""
    run_command(PIP_INSTALL + ["--upgrade", "pip", "wheel"])


def install_requirements() -> None:
    """Install requirements."""
    print("Installing requirements...")
    cmd = PIP_INSTALL + ["-r", str(REQUIREMENTS_FILE)]
    run_command(cmd)


def init_jupyter_extension() -> None:
    """Initialize the Jupyter extension."""
    jlpm_command = shutil.which("jlpm")
    if not jlpm_command:  # in venv?
        jlpm_exe = "jlpm" + (".exe" if sys.platform == "win32" else "")
        jlpm_path = BIN_DIR / jlpm_exe
        if jlpm_path.is_file() and os.access(jlpm_path, os.X_OK):
            jlpm_command = str(jlpm_path)
    if not jlpm_command:
        print("jlpm is not found as a command. Please install it first.")
        return
    run_command([jlpm_command, "install"], cwd=str(ROOT_DIR))
    run_command([jlpm_command, "build"], cwd=str(ROOT_DIR))
    run_command(PIP_INSTALL + ["-e", "."], cwd=str(ROOT_DIR))
    _cmd = ["labextension", "develop", "--overwrite", "."]
    extension_cmd = [JUPYTER_COMMAND] + _cmd
    run_command(extension_cmd, cwd=str(ROOT_DIR))


def done() -> None:
    """Print the done message."""
    dev_run_path = os.path.join("scripts", "dev", "run.py")
    msg = (
        "Done. \n"
        "You can run 'make dev' to start jupyter-lab with the "
        "installed extension.\n"
        "If make is not available, you can run \n"
        f"`python {dev_run_path}` instead."
    )
    print(msg)


def main() -> None:
    """Initialize this project."""
    create_venv()
    pip_upgrade_base()
    install_requirements()
    init_jupyter_extension()
    done()


if __name__ == "__main__":
    main()



"""Handle file uploads."""

import json
import os
from pathlib import Path

from jupyter_server.base.handlers import APIHandler
from pathvalidate import sanitize_filename
from tornado.web import HTTPError

ALLOWED_EXTENSIONS = {
    ".txt",
    ".pdf",
    ".doc",
    ".docx",
    ".rtf",
    ".xlsx",
    ".xls",
    ".csv",
    ".json",
    ".yaml",
    ".yml",
    ".xml",
    ".md",
    ".odt",
}


class UploadHandler(APIHandler):
    """Upload handler to handle file uploads."""

    def data_received(self, chunk: bytes) -> None:
        """Just to make the linter happy.

        avoid: Pylint W0223:abstract-method

        Parameters
        ----------
        chunk : bytes
            The chunk of data received.
        """
        super().data_received(chunk)  # pragma: no cover

    async def post(self) -> None:
        """Handle a POST request.

        Raises
        ------
        HTTPError
            If the request data is invalid.

        The request should contain the file to upload.

        Example URL:
        /waldiez/upload
        """
        file_info = self.request.files.get("file", None)
        if not file_info:
            raise HTTPError(400, reason="No file in request")
        file = file_info[0]
        # make sure the filename is safe (no extra dots, slashes, etc.)
        filename = sanitize_filename(file["filename"])
        # make sure the file extension is allowed
        if not is_allowed_extension(filename):
            raise HTTPError(400, reason="File extension not allowed")
        # save the file
        file_path = self._get_file_path(filename)
        with open(file_path, "wb") as file_obj:
            file_obj.write(file["body"])
        self.finish(json.dumps({"path": str(file_path)}))

    def _get_file_path(self, file_name: str) -> Path:
        """Get the actual path of the file.

        Parameters
        ----------
        file_name : str
            The filename to save.

        Returns
        -------
        str
            The actual path of the file (where it will be saved).
        """
        joined = os.path.join(self.contents_manager.root_dir, file_name)
        return Path(os.path.abspath(joined))


def is_allowed_extension(filename: str) -> bool:
    """Check if the file extension is allowed.

    Parameters
    ----------
    filename : str
        The filename to check.

    Returns
    -------
    bool
        True if the extension is allowed, False otherwise.
    """
    return any(filename.endswith(extension) for extension in ALLOWED_EXTENSIONS)



"""Build container image."""

import argparse
import os
import shutil
import subprocess  # nosemgrep # nosec
import sys
from pathlib import Path
from typing import List

try:
    from dotenv import load_dotenv
except ImportError:
    pass
else:
    load_dotenv(override=True)


_ROOT_DIR = Path(__file__).parent.parent.parent.resolve()
_DEFAULT_IMAGE = os.environ.get("IMAGE_NAME", "waldiez/jupyter")
_FALLBACK_TAG = "dev" if "--dev" in sys.argv else "latest"
_DEFAULT_TAG = os.environ.get("IMAGE_TAG", _FALLBACK_TAG)
_DEFAULT_PLATFORM = os.environ.get("PLATFORM", "linux/amd64")


def get_container_cmd() -> str:
    """Get the container command to use.

    Returns
    -------
    str
        The container command to use. Either "docker" or "podman".

    Raises
    ------
    RuntimeError
        If neither "docker" nor "podman" is found.
    """
    from_env = os.environ.get("CONTAINER_COMMAND", "")
    if from_env and from_env in ["docker", "podman"]:
        return from_env
    # prefer podman over docker if found
    if shutil.which("podman"):
        return "podman"
    if not shutil.which("docker"):
        raise RuntimeError("Could not find docker or podman.")
    return "docker"


def run_command(command: List[str]) -> None:
    """Run a command.

    Parameters
    ----------
    command : List[str]
        The command to run.

    Raises
    ------
    subprocess.CalledProcessError
        If the command returns a non-zero exit status.
    """
    print(f"Running command: \n{' '.join(command)}\n")
    subprocess.run(
        command,
        check=True,
        env=os.environ,
        cwd=_ROOT_DIR,
        stdout=sys.stdout,
        stderr=sys.stderr,
        text=True,
    )  # nosemgrep # nosec


# pylint: disable=too-many-positional-arguments
def build_image(
    container_file: str,
    image_name: str,
    image_tag: str,
    platform: str,
    container_command: str,
    no_cache: bool,
    build_args: List[str],
) -> None:
    """Build the container image.

    Parameters
    ----------
    container_file : str
        Path to the Containerfile.
    image_name : str
        Name of the image to build.
    image_tag : str
        Tag of the image to build.
    platform : str
        Set platform if the image is multi-platform.
    container_command : str
        The container command to use.
    no_cache : bool
        Do not use cache when building the image.
    build_args : List[str]
        Build arguments.

    Raises
    ------
    subprocess.CalledProcessError
        If the command returns a non-zero exit status.
    """
    cmd = [
        container_command,
        "build",
        "--platform",
        platform,
        "--tag",
        f"{image_name}:{image_tag}",
        "-f",
        container_file,
    ]
    if no_cache:
        cmd.append("--no-cache")
    for arg in build_args:
        cmd.extend(["--build-arg", arg])
    if container_command == "docker":
        cmd.extend(["--progress=plain"])
    cmd.append(".")
    run_command(cmd)


def cli() -> argparse.ArgumentParser:
    """Create the CLI parser.

    Returns
    -------
    argparse.ArgumentParser
        The CLI parser.
    """
    parser = argparse.ArgumentParser(description="Build container image.")
    parser.add_argument(
        "--image-name",
        type=str,
        default=_DEFAULT_IMAGE,
        help="Name of the image to build.",
    )
    parser.add_argument(
        "--image-tag",
        type=str,
        default=_DEFAULT_TAG,
        help="Tag of the image to build.",
    )
    parser.add_argument(
        "--platform",
        type=str,
        default=_DEFAULT_PLATFORM,
        choices=["linux/amd64", "linux/arm64"],
        help="Set platform if the image is multi-platform.",
    )
    parser.add_argument(
        "--container-command",
        default=get_container_cmd(),
        choices=["docker", "podman"],
        help="The container command to use.",
    )
    parser.add_argument(
        "--build-args",
        type=str,
        action="append",
        help="Build arguments.",
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Do not use cache when building the image.",
    )
    parser.add_argument(
        "--dev",
        action="store_true",
        help="Use development container.",
    )
    return parser


def main() -> None:
    """Parse the CLI arguments and build the container image.

    Raises
    ------
    subprocess.CalledProcessError
        If the command returns a non-zero exit status
    """
    args = cli().parse_args()
    build_args = args.build_args or []
    container_file = "Containerfile" if not args.dev else "Containerfile.dev"
    build_image(
        container_file,
        args.image_name,
        args.image_tag,
        args.platform,
        args.container_command,
        args.no_cache,
        build_args,
    )


if __name__ == "__main__":
    main()



"""Pytest configuration file for the Waldiez Jupyter extension."""

import json
import os
from pathlib import Path
from typing import Any, Dict, Generator, List

import pytest
from filelock import FileLock
from jupyter_server import DEFAULT_STATIC_FILES_PATH

os.environ["JUPYTER_PLATFORM_DIRS"] = "1"
os.environ["JUPYTER_IS_TESTING"] = "1"

pytest_plugins = ("pytest_jupyter.jupyter_server",)

# pylint: disable=unused-argument,redefined-outer-name


@pytest.fixture
def jp_server_config(jp_server_config: Dict[str, Any]) -> Dict[str, Any]:
    """Add the extension to the server configuration.

    Parameters
    ----------
    jp_server_config : Dict[str, Any]
        The Jupyter server configuration.
    Returns
    -------
    Dict[str, Any]
        The updated server configuration.
    """
    return {"ServerApp": {"jpserver_extensions": {"waldiez_jupyter": True}}}


def _load_worker_list(fn: Path) -> List[str]:
    """Read the worker id list from a file."""
    return json.loads(fn.read_text())["workers"]


def _write_worker_list(fn: Path, workers: List[str]) -> None:
    """Write the worker id list to a file."""
    fn.write_text(json.dumps({"workers": workers}))


def _remove_monaco_latest_version() -> None:
    """Remove the monaco latest version file."""
    # remove STATIC_FILES_PATH/monaco_latest_version if it exists
    # to force/check for a new download
    static_path = Path(DEFAULT_STATIC_FILES_PATH)
    monaco_latest_version = static_path / "monaco_latest_version"
    if monaco_latest_version.exists():
        monaco_latest_version.unlink(missing_ok=True)


def _before(worker_id: str) -> None:
    """Run before all tests."""
    if worker_id == "master":
        # not executing with multiple workers
        _remove_monaco_latest_version()
        return
    # credits:
    # https://github.com/pytest-dev/pytest-xdist/issues/783#issuecomment-1793593178
    lock_file = Path(DEFAULT_STATIC_FILES_PATH) / "workers.json.lock"
    worker_list_file = Path(DEFAULT_STATIC_FILES_PATH) / "workers.json"
    with FileLock(str(lock_file)):
        is_first_worker = not worker_list_file.is_file()
        if worker_list_file.is_file():
            workers = _load_worker_list(worker_list_file)
        else:
            workers = []
            is_first_worker = True
        workers.append(worker_id)
        _write_worker_list(worker_list_file, workers)
        if is_first_worker:
            _remove_monaco_latest_version()


def _after() -> None:
    """Run after all tests."""


# pylint: disable=unused-argument
@pytest.fixture(scope="session", autouse=True)
def before_and_after_tests(
    request: Any, tmp_path_factory: pytest.TempPathFactory, worker_id: str
) -> Generator[None, None, None]:
    """Fixture to run before and after all tests.

    Parameters
    ----------
    request : Any
        The request object.
    tmp_path_factory : pytest.TempPathFactory
        The temporary path factory.
    worker_id : str
        The worker ID.

    Yields
    ------
    Generator[None, None, None]
        The generator to run the tests.
    """
    _before(worker_id)
    try:
        yield
    finally:
        _after()


@pytest.fixture
def data_dir() -> Path:
    """Return the tests data directory.

    Returns
    -------
    Path
        The tests data directory.
    """
    here = Path(__file__).parent
    return here / "waldiez_jupyter" / "tests" / "data"



"""Extension handlers for the Jupyter Server."""

from jupyter_server.serverapp import ServerWebApplication
from jupyter_server.utils import url_path_join

from .extra_static_files import ensure_extra_static_files
from .files_handler import FilesHandler
from .upload_handler import UploadHandler


def setup_handlers(web_app: ServerWebApplication) -> None:
    """Add the extension handlers to the Jupyter server web application.

    Parameters
    ----------
    web_app : ServerWebApplication
        Jupyter server web application
    """
    host_pattern = ".*$"
    static_path_or_paths = web_app.settings["static_path"]
    if isinstance(static_path_or_paths, list):
        static_path = static_path_or_paths[-1]
    else:
        static_path = static_path_or_paths
    ensure_extra_static_files(static_path)
    base_url = web_app.settings["base_url"]
    files_pattern = url_path_join(base_url, "waldiez", "files")
    upload_pattern = url_path_join(base_url, "waldiez", "upload")
    web_app.add_handlers(
        host_pattern,
        [(files_pattern, FilesHandler), (upload_pattern, UploadHandler)],
    )



"""Cleanup."""

# pylint: disable=duplicate-code,broad-except
import glob
import os
import shutil
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent.parent

DIR_PATTERNS = [
    "__pycache__",
    ".cache",
    ".pytest_cache",
    ".mypy_cache",
    ".ruff_cache",
    "*.egg-info",
    "htmlcov",
]

FILE_PATTERNS = [
    "*.pyc",
    "*.pyo",
    "*.pyc~",
    "*.py~",
    "*~",
    ".*~",
    ".DS_Store",
    "._DS_Store",
    "._.DS_Store",
    ".coverage*",
]

ROOT_LEVEL_DIRS = [
    "reports",
]


def _remove_dirs() -> None:
    for pattern in DIR_PATTERNS:
        for dirpath in glob.glob(f"./**/{pattern}", recursive=True):
            sys.stdout.write(f"removing {dirpath}\n")
            try:
                shutil.rmtree(dirpath)
            except BaseException:
                pass
    for dirname in ROOT_LEVEL_DIRS:
        directory = os.path.join(ROOT_DIR, dirname)
        if os.path.isdir(directory):
            sys.stdout.write(f"removing {dirname}\n")
            try:
                shutil.rmtree(directory)
            except BaseException:
                pass


def _remove_files() -> None:
    for pattern in FILE_PATTERNS:
        for filepath in glob.glob(f"./**/{pattern}", recursive=True):
            sys.stdout.write(f"removing {filepath}\n")
            os.remove(filepath)


def main() -> None:
    """Cleanup unnecessary files and directories."""
    _cwd = os.getcwd()
    os.chdir(ROOT_DIR)
    _remove_dirs()
    _remove_files()
    if os.getcwd() != _cwd:
        os.chdir(_cwd)


if __name__ == "__main__":
    main()



"""Start/stop dev mode.

--react (frontend)
if both, we subprocess two calls
"""

import multiprocessing
import os
import subprocess
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).resolve().parent.parent.parent

VENV_DIR = ROOT_DIR / ".venv"
EXE_ = ".exe" if sys.platform == "win32" else ""
BIN_DIR = VENV_DIR / "Scripts" if sys.platform == "win32" else VENV_DIR / "bin"
JLPM_COMMAND = str(BIN_DIR / f"jlpm{EXE_}")
JUPYTER_COMMAND = str(BIN_DIR / f"jupyter{EXE_}")
PYTHON_COMMAND = str(BIN_DIR / f"python{EXE_}")


def activate_venv() -> None:
    """Activate the virtual environment."""
    if not BIN_DIR.is_dir() or not (BIN_DIR / f"python{EXE_}").is_file():
        return
    # pylint: disable=inconsistent-quotes
    os.environ["PATH"] = f"{BIN_DIR}{os.pathsep}{os.environ['PATH']}"

    if sys.platform == "win32":
        os.environ["VIRTUAL_ENV"] = str(VENV_DIR)
        os.environ["PROMPT"] = f"({VENV_DIR.name}) $P$G"


def _do_start_jupyter() -> None:
    os.environ["PYTHONUNBUFFERED"] = "1"
    cmd = [JUPYTER_COMMAND, "lab", "-y", "--no-browser", "--autoreload"]
    if (ROOT_DIR / "examples").exists():
        cmd += ["--notebook-dir", "examples"]
    subprocess.run(  # nosemgrep  # nosec
        cmd,
        cwd=ROOT_DIR,
        check=True,
        env=os.environ,
    )


def _start_jupyter() -> None:
    """Start the Jupyter server."""
    pid_path = ROOT_DIR / "jupyter.pid"
    if pid_path.exists():
        print("Jupyter server already running?, or stale pid file")
        return
    multiprocessing.set_start_method("spawn")
    p = multiprocessing.Process(target=_do_start_jupyter, daemon=True)
    p.start()
    pid = p.pid
    with open(ROOT_DIR / "jupyter.pid", "w", encoding="utf-8") as f:
        f.write(str(pid))


def _start_jlpm() -> None:
    try:
        # pylint: disable=subprocess-run-check, line-too-long
        subprocess.run([JLPM_COMMAND, "run", "watch"], cwd=ROOT_DIR, check=True)  # nosemgrep  # nosec
    except KeyboardInterrupt:
        _stop()


def _stop_using_pid() -> None:
    """Stop the Jupyter server using the PID."""
    pid_path = ROOT_DIR / "jupyter.pid"
    if not pid_path.exists():
        return
    pid = None
    with open(pid_path, "r", encoding="utf-8") as f:
        try:
            pid = int(f.read())
        except BaseException as e:  # pylint: disable=broad-except
            print(f"Error reading pid: {e}")
    if pid is None:
        return
    try:
        os.kill(pid, 15)
    except BaseException:  # pylint: disable=broad-except
        pass
    finally:
        pid_path.unlink()


def _stop_using_tasklist() -> None:
    """Stop the Jupyter server using tasklist."""
    cmd = 'tasklist /fi "imagename eq jupyter-lab.exe" /fo csv /nh'  # noqa
    result = subprocess.run(  # nosemgrep  # nosec
        cmd,
        shell=True,
        check=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if not result.stdout:
        return
    try:
        pid = int(result.stdout.decode().split(",")[1].strip('"'))
    except BaseException:  # pylint: disable=broad-except
        return
    try:
        os.kill(pid, 15)
    except BaseException:  # pylint: disable=broad-except
        pass


def _stop() -> None:
    """Stop the Jupyter server."""
    _stop_using_pid()
    # in case no pid file was found
    if sys.platform == "win32":
        _stop_using_tasklist()
        return
    for kwd in ("ipykernel_launcher", "jupyter-lab"):
        # pylint: disable=line-too-long
        cmd = f"ps aux | grep -v grep | grep {kwd} | awk '{{print $2}}' | xargs kill -9 > /dev/null 2>&1 || true"
        subprocess.run(
            cmd,
            shell=True,
            check=True,
        )  # nosemgrep  # nosec


def main() -> None:
    """Run the main script."""
    if "--stop" in sys.argv:
        _stop()
        return
    _start_jupyter()
    _start_jlpm()


if __name__ == "__main__":
    activate_venv()
    main()



"""Ensure extra static files (for monaco editor) are present.

We can serve these files from the Jupyter server or not.
The option is available on the plugin configuration page.
The default is to do serve them. If not, the files are served
from the default CDN that monaco loader uses.
"""

import hashlib
import io
import json
import logging
import os
import shutil
import tarfile
from pathlib import Path
from typing import Tuple, Union

import urllib3

REGISTRY_BASE_URL = "https://registry.npmjs.org"
PACKAGE_NAME = "monaco-editor"
LOG = logging.getLogger(__name__)

# pylint: disable=broad-except


def ensure_extra_static_files(static_root_path: Union[str, Path]) -> None:
    """Ensure extra static files are present.

    Parameters
    ----------
    static_root_path : Union[str, Path]
        The path to the static files directory

    Raises
    ------
    RuntimeError
        If the monaco editor files are not found.
    """
    if not isinstance(static_root_path, Path):
        static_root_path = Path(static_root_path)
    static_root_path.mkdir(parents=True, exist_ok=True)
    try:
        details = _get_package_details()
    except BaseException as error:
        LOG.error("Failed to get the latest version of monaco: %s", error)
        return
    version_file = static_root_path / "monaco_latest_version"
    if version_file.exists():
        with open(version_file, "r", encoding="utf-8") as file:
            current_version = file.read()
    else:
        current_version = None
    loader_js = static_root_path / "vs" / "loader.js"
    if current_version != details[0] or not loader_js.exists():
        LOG.info("Downloading monaco editor files...")
        try:
            _download_monaco_editor(details, static_root_path)
        except BaseException as error:
            LOG.error("Failed to download monaco editor files: %s", error)
            return
        with open(version_file, "w", encoding="utf-8") as file:
            file.write(details[0])
    if not loader_js.exists():
        LOG.error("Monaco editor files not found.")
        LOG.error("Path: %s", static_root_path)
        LOG.error("Files: %s", os.listdir(static_root_path))
        LOG.error("vs folder: %s", os.listdir(static_root_path / "vs"))
        raise RuntimeError("Failed to download monaco editor files.")
    LOG.info("Monaco editor files are up-to-date.")


def _get_package_details() -> Tuple[str, str, str]:
    """Get details about the latest version of monaco editor.

    Returns
    -------
    Tuple[str, str, str]
        The latest version, download url, and SHA-1 sum.
    """
    http = urllib3.PoolManager()
    response = http.request("GET", f"{REGISTRY_BASE_URL}/{PACKAGE_NAME}")
    data = json.loads(response.data)
    latest_version = data["dist-tags"]["latest"]
    latest_version_data = data["versions"][latest_version]
    url = latest_version_data["dist"]["tarball"]
    sha_sum = latest_version_data["dist"]["shasum"]
    return latest_version, url, sha_sum


def _download_monaco_editor(
    details: Tuple[str, str, str],
    static_root_path: Union[str, Path],
) -> None:
    """Download and extract the monaco editor files.

    Parameters
    ----------
    details : Tuple[str, str, str]
        The version, download url, and SHA-1 sum.

    static_root_path : Union[str, Path]
        The path to the static files directory.
    """
    http = urllib3.PoolManager()
    version_url, version_sha_sum = details[1], details[2]
    response = http.request("GET", version_url)
    sha_sum = hashlib.sha1(response.data, usedforsecurity=False).hexdigest()
    if sha_sum != version_sha_sum:
        raise ValueError("SHA-1 sum mismatch.")
    os.makedirs(static_root_path, exist_ok=True)
    with tarfile.open(fileobj=io.BytesIO(response.data)) as tar:
        tar.extractall(static_root_path)  # nosemgrep # nosec
    monaco_editor_path = os.path.join(static_root_path, "package")
    src_dir = os.path.join(monaco_editor_path, "min", "vs")
    if not os.path.exists(src_dir):
        raise FileNotFoundError("Failed to extract monaco editor files.")
    dst_dir = os.path.join(static_root_path, "vs")
    if os.path.exists(dst_dir):
        shutil.rmtree(dst_dir)
    shutil.move(src_dir, static_root_path)
    shutil.rmtree(monaco_editor_path)



"""Tests for the upload handler."""

import json
from pathlib import Path
from typing import Any, Callable
from uuid import uuid4

import pytest
import tornado


async def test_valid_file_upload(
    jp_fetch: Callable[..., Any],
    jp_root_dir: Path,
    data_dir: Path,
) -> None:
    """Test a valid file upload.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    jp_root_dir : Path
        The Jupyter server root directory.
    data_dir : Path
        The data directory.
    """
    file_path = data_dir / "dummy.txt"
    boundary = uuid4().hex
    headers = {
        "Content-Type": f"multipart/form-data; boundary={boundary}",
    }
    filename = file_path.name
    body = (
        f"--{boundary}\r\n"
        'Content-Disposition: form-data; name="file"; '
        f'filename="{filename}"\r\n'
        "Content-Type: text/plain\r\n\r\n"
        f"{file_path.read_text()}\r\n"
        f"--{boundary}--\r\n"
    )
    response = await jp_fetch(
        "waldiez",
        "upload",
        method="POST",
        headers=headers,
        body=body,
    )
    assert response.code == 200
    response_data = json.loads(response.body.decode("utf-8"))
    assert "path" in response_data
    destination = jp_root_dir / file_path.name
    assert response_data["path"] == str(destination)
    assert destination.exists()
    destination.unlink()


async def test_no_file_upload(jp_fetch: Callable[..., Any]) -> None:
    """Test an upload request without a file.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    """
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch("waldiez", "upload", method="POST", body="")
    assert exc_info.value.code == 400


async def test_invalid_file_upload(
    jp_fetch: Callable[..., Any],
    data_dir: Path,
) -> None:
    """Test an invalid file upload.

    Parameters
    ----------
    jp_fetch : Callable[..., Any]
        The Jupyter server fetch function.
    data_dir : Path
        The data directory.
    """
    image = data_dir / "blank.png"
    boundary = uuid4().hex
    headers = {
        "Content-Type": f"multipart/form-data; boundary={boundary}",
    }
    image_data = image.read_bytes()
    body = (
        f"--{boundary}\r\n"
        'Content-Disposition: form-data; name="file"; '
        'filename="blank.png"\r\n'
        "Content-Type: image/png\r\n\r\n"
    ).encode("utf-8")
    body += image_data
    body += f"\r\n--{boundary}--\r\n".encode("utf-8")
    with pytest.raises(tornado.httpclient.HTTPClientError) as exc_info:
        await jp_fetch(
            "waldiez",
            "upload",
            method="POST",
            headers=headers,
            body=body,
        )
    assert exc_info.value.code == 400
    assert exc_info.value.response
    assert "File extension not allowed" in exc_info.value.response.reason
